@@ -10,8 +10,11 @@ from hatch.config.user import ConfigFile, RootConfig
 from hatch.project.core import Project
 from hatch.utils.fs import Path
 from hatch.utils.platform import Platform
+from hatch.utils.runner import ExecutionContext
 
 if TYPE_CHECKING:
+    from collections.abc import Generator
+
     from packaging.requirements import Requirement
 
     from hatch.env.plugin.interface import EnvironmentInterface
@@ -42,23 +45,42 @@ class Application(Terminal):
     def config(self) -> RootConfig:
         return self.config_file.model
 
+    def expand_environments(self, env_name: str) -> list[str]:
+        if env_name in self.project.config.internal_matrices:
+            return list(self.project.config.internal_matrices[env_name]['envs'])
+
+        if env_name in self.project.config.matrices:
+            return list(self.project.config.matrices[env_name]['envs'])
+
+        if env_name in self.project.config.internal_envs:
+            return [env_name]
+
+        if env_name in self.project.config.envs:
+            return [env_name]
+
+        return []
+
     def get_environment(self, env_name: str | None = None) -> EnvironmentInterface:
         if env_name is None:
             env_name = self.env
 
         if env_name in self.project.config.internal_envs:
             config = self.project.config.internal_envs[env_name]
-        elif env_name not in self.project.config.envs:
-            self.abort(f'Unknown environment: {env_name}')
-        else:
+        elif env_name in self.project.config.envs:
             config = self.project.config.envs[env_name]
+        else:
+            self.abort(f'Unknown environment: {env_name}')
 
         environment_type = config['type']
         environment_class = self.plugins.environment.get(environment_type)
         if environment_class is None:
             self.abort(f'Environment `{env_name}` has unknown type: {environment_type}')
 
-        if is_isolated_environment(env_name, config):
+        from hatch.env.internal import is_isolated_environment
+
+        if self.project.location.is_file():
+            data_directory = isolated_data_directory = self.data_dir / 'env' / environment_type / '.scripts'
+        elif is_isolated_environment(env_name, config):
             data_directory = isolated_data_directory = self.data_dir / 'env' / '.internal' / env_name
         else:
             data_directory = self.get_env_directory(environment_type)
@@ -76,7 +98,7 @@ class Application(Terminal):
             isolated_data_directory,
             self.platform,
             self.verbosity,
-            self.get_safe_application(),
+            self,
         )
 
     # Ensure that this method is clearly written since it is
@@ -85,79 +107,133 @@ class Application(Terminal):
         if not environment.exists():
             self.env_metadata.reset(environment)
 
-            with self.status(f'Creating environment: {environment.name}'):
+            with environment.app_status_creation():
                 environment.create()
 
             if not environment.skip_install:
                 if environment.pre_install_commands:
-                    with self.status('Running pre-installation commands'):
-                        self.run_shell_commands(environment, environment.pre_install_commands, source='pre-install')
+                    with environment.app_status_pre_installation():
+                        self.run_shell_commands(
+                            ExecutionContext(
+                                environment,
+                                shell_commands=environment.pre_install_commands,
+                                source='pre-install',
+                                show_code_on_error=True,
+                            )
+                        )
 
-                if environment.dev_mode:
-                    with self.status('Installing project in development mode'):
+                with environment.app_status_project_installation():
+                    if environment.dev_mode:
                         environment.install_project_dev_mode()
-                else:
-                    with self.status('Installing project'):
+                    else:
                         environment.install_project()
 
                 if environment.post_install_commands:
-                    with self.status('Running post-installation commands'):
-                        self.run_shell_commands(environment, environment.post_install_commands, source='post-install')
+                    with environment.app_status_post_installation():
+                        self.run_shell_commands(
+                            ExecutionContext(
+                                environment,
+                                shell_commands=environment.post_install_commands,
+                                source='post-install',
+                                show_code_on_error=True,
+                            )
+                        )
+
+        with environment.app_status_dependency_state_check():
+            new_dep_hash = environment.dependency_hash()
 
-        new_dep_hash = environment.dependency_hash()
         current_dep_hash = self.env_metadata.dependency_hash(environment)
         if new_dep_hash != current_dep_hash:
-            with self.status('Checking dependencies'):
+            with environment.app_status_dependency_installation_check():
                 dependencies_in_sync = environment.dependencies_in_sync()
 
             if not dependencies_in_sync:
-                with self.status('Syncing dependencies'):
+                with environment.app_status_dependency_synchronization():
                     environment.sync_dependencies()
                     new_dep_hash = environment.dependency_hash()
 
             self.env_metadata.update_dependency_hash(environment, new_dep_hash)
 
-    def run_shell_commands(
-        self,
-        environment: EnvironmentInterface,
-        commands: list[str],
-        source='cmd',
-        *,
-        force_continue=False,
-        show_code_on_error=True,
-    ):
-        with environment.command_context():
+    def run_shell_commands(self, context: ExecutionContext) -> None:
+        with context.env.command_context():
             try:
-                resolved_commands = list(environment.resolve_commands(commands))
+                resolved_commands = list(context.env.resolve_commands(context.shell_commands))
             except Exception as e:  # noqa: BLE001
                 self.abort(str(e))
 
             first_error_code = None
-            should_display_command = self.verbose or len(resolved_commands) > 1
+            should_display_command = not context.hide_commands and (self.verbose or len(resolved_commands) > 1)
             for i, raw_command in enumerate(resolved_commands, 1):
                 if should_display_command:
-                    self.display(f'{source} [{i}] | {raw_command}')
+                    self.display(f'{context.source} [{i}] | {raw_command}')
 
                 command = raw_command
-                continue_on_error = force_continue
+                continue_on_error = context.force_continue
                 if raw_command.startswith('- '):
                     continue_on_error = True
                     command = command[2:]
 
-                process = environment.run_shell_command(command)
+                process = context.env.run_shell_command(command)
                 if process.returncode:
                     first_error_code = first_error_code or process.returncode
                     if continue_on_error:
                         continue
 
-                    if show_code_on_error:
+                    if context.show_code_on_error:
                         self.abort(f'Failed with exit code: {process.returncode}', code=process.returncode)
                     else:
                         self.abort(code=process.returncode)
 
-            if first_error_code and force_continue:
+            if first_error_code and context.force_continue:
                 self.abort(code=first_error_code)
 
+    def runner_context(
+        self,
+        environments: list[str],
+        *,
+        ignore_compat: bool = False,
+        display_header: bool = False,
+    ) -> Generator[ExecutionContext, None, None]:
+        from hatch.utils.structures import EnvVars
+
+        if self.verbose or len(environments) > 1:
+            display_header = True
+
+        any_compatible = False
+        incompatible = {}
+        with self.project.ensure_cwd():
+            for env_name in environments:
+                environment = self.get_environment(env_name)
+                if not environment.exists():
+                    try:
+                        environment.check_compatibility()
+                    except Exception as e:  # noqa: BLE001
+                        if ignore_compat:
+                            incompatible[environment.name] = str(e)
+                            continue
+
+                        self.abort(f'Environment `{env_name}` is incompatible: {e}')
+
+                any_compatible = True
+                if display_header:
+                    self.display_header(environment.name)
+
+                context = ExecutionContext(environment)
+                yield context
+
+                self.prepare_environment(environment)
+                with EnvVars(context.env_vars):
+                    self.run_shell_commands(context)
+
+        if incompatible:
+            num_incompatible = len(incompatible)
+            padding = '\n' if any_compatible else ''
+            self.display_warning(
+                f'{padding}Skipped {num_incompatible} incompatible environment{"s" if num_incompatible > 1 else ""}:'
+            )
+            for env_name, reason in incompatible.items():
+                self.display_warning(f'{env_name} -> {reason}')
+
     def ensure_environment_plugin_dependencies(self) -> None:
         self.ensure_plugin_dependencies(
             self.project.config.env_requires_complex, wait_message='Syncing environment plugin requirements'
@@ -219,60 +295,19 @@ class Application(Terminal):
 
     @cached_property
     def shell_data(self) -> tuple[str, str]:
-        import shellingham
+        from hatch.utils.shells import detect_shell
 
-        try:
-            return shellingham.detect_shell()
-        except shellingham.ShellDetectionFailure:
-            path = self.platform.default_shell
-            return Path(path).stem, path
+        return detect_shell(self.platform)
 
     @cached_property
     def env_metadata(self) -> EnvironmentMetadata:
         return EnvironmentMetadata(self.data_dir / 'env' / '.metadata', self.project.location)
 
-    @staticmethod
-    def is_internal_default_environment(env_name: str, config: dict[str, Any]) -> bool:
-        from hatch.env.internal import get_internal_env_config
-
-        internal_config = get_internal_env_config().get(env_name)
-        if not internal_config:
-            return False
-
-        # Only consider things that would modify the actual installation, other options like extra scripts don't matter
-        for key in ('dependencies', 'extra-dependencies', 'features'):
-            if config.get(key) != internal_config.get(key):
-                return False
-
-        return True
-
     def abort(self, text='', code=1, **kwargs):
         if text:
             self.display_error(text, **kwargs)
         self.__exit_func(code)
 
-    def get_safe_application(self) -> SafeApplication:
-        return SafeApplication(self)
-
-
-class SafeApplication:
-    def __init__(self, app: Application):
-        self.abort = app.abort
-        self.display = app.display
-        self.display_critical = app.display_critical
-        self.display_info = app.display_info
-        self.display_error = app.display_error
-        self.display_success = app.display_success
-        self.display_waiting = app.display_waiting
-        self.display_warning = app.display_warning
-        self.display_debug = app.display_debug
-        self.display_mini_header = app.display_mini_header
-        # Divergence from what the backend provides
-        self.prompt = app.prompt
-        self.confirm = app.confirm
-        self.status = app.status
-        self.status_if = app.status_if
-
 
 class EnvironmentMetadata:
     def __init__(self, data_dir: Path, project_path: Path):
@@ -307,6 +342,8 @@ class EnvironmentMetadata:
         metadata_file.write_text(json.dumps(metadata))
 
     def _metadata_file(self, environment: EnvironmentInterface) -> Path:
+        from hatch.env.internal import is_isolated_environment
+
         if is_isolated_environment(environment.name, environment.config):
             return self.__data_dir / '.internal' / f'{environment.name}.json'
 
@@ -314,34 +351,4 @@ class EnvironmentMetadata:
 
     @cached_property
     def _storage_dir(self) -> Path:
-        from base64 import urlsafe_b64encode
-        from hashlib import sha256
-
-        return self.__data_dir / urlsafe_b64encode(sha256(str(self.__project_path).encode()).digest())[:8].decode()
-
-
-def is_isolated_environment(env_name: str, config: dict[str, Any]) -> bool:
-    # Provide super isolation and immunity to project-level environment removal only when the environment:
-    #
-    # 1. Does not require the project being installed
-    # 2. The default configuration is used
-    #
-    # For example, the environment for static analysis depends only on Ruff at a specific default
-    # version. This environment does not require the project and can be reused by every project to
-    # improve responsiveness. However, if the user for some reason chooses to override the dependencies
-    # to use a different version of Ruff, then the project would get its own environment.
-    if not config.get('skip-install', False):
-        return False
-
-    from hatch.env.internal import get_internal_env_config
-
-    internal_config = get_internal_env_config().get(env_name)
-    if not internal_config:
-        return False
-
-    # Only consider things that would modify the actual installation, other options like extra scripts don't matter
-    for key in ('dependencies', 'extra-dependencies', 'features'):
-        if config.get(key) != internal_config.get(key):
-            return False
-
-    return True
+        return self.__data_dir / self.__project_path.id
