@@ -22,5 +22,14 @@ class PublishEnvVars:
     USER = 'HATCH_INDEX_USER'
     AUTH = 'HATCH_INDEX_AUTH'
     REPO = 'HATCH_INDEX_REPO'
+    CA_CERT = 'HATCH_INDEX_CA_CERT'
+    CLIENT_CERT = 'HATCH_INDEX_CLIENT_CERT'
+    CLIENT_KEY = 'HATCH_INDEX_CLIENT_KEY'
     PUBLISHER = 'HATCH_PUBLISHER'
     OPTIONS = 'HATCH_PUBLISHER_OPTIONS'
+
+
+class PythonEnvVars:
+    CUSTOM_SOURCE_PREFIX = 'HATCH_PYTHON_CUSTOM_SOURCE_'
+    CUSTOM_PATH_PREFIX = 'HATCH_PYTHON_CUSTOM_PATH_'
+    CUSTOM_VERSION_PREFIX = 'HATCH_PYTHON_CUSTOM_VERSION_'
